////////////////////////////////////////////////////////////////////////////
//
//      Name:             calc.cpp
//      Erstellt:         21.05.2005
//      Letzte nderung:  12.12.2005
//
//      Beschreibung:     Klasse zum Berechnen der neuen DWZ
//
////////////////////////////////////////////////////////////////////////////

#include "main.h"

#include <windows.h>
#include <stdio.h>
#include <fstream>
#include <math.h>
using namespace std;

#include "calc.h"

// ---------------------------------------------------------------------- \\

// Tabelle fr Gewinnwahrscheinlichkeit
CALC_ROW table [] = {
{0,3,0.5,0.5},
{4,10,0.51,0.49},
{11,17,0.52,0.48},
{18,25,0.53,0.47},
{26,32,0.54,0.46},
{33,39,0.55,0.45},
{40,46,0.56,0.44},
{47,53,0.57,0.43},
{54,61,0.58,0.42},
{62,68,0.59,0.41},
{69,76,0.6,0.4},
{77,83,0.61,0.39},
{84,91,0.62,0.38},
{92,98,0.63,0.37},
{99,106,0.64,0.36},
{107,113,0.65,0.35},
{114,121,0.66,0.34},
{122,129,0.67,0.33},
{130,137,0.68,0.32},
{138,145,0.69,0.31},
{146,153,0.7,0.3},
{154,162,0.71,0.29},
{163,170,0.72,0.28},
{171,179,0.73,0.27},
{180,188,0.74,0.26},
{189,197,0.75,0.25},
{198,206,0.76,0.24},
{207,215,0.77,0.23},
{216,225,0.78,0.22},
{226,235,0.79,0.21},
{236,245,0.8,0.2},
{246,256,0.81,0.19},
{257,267,0.82,0.18},
{268,278,0.83,0.17},
{279,290,0.84,0.16},
{291,302,0.85,0.15},
{303,315,0.86,0.14},
{316,328,0.87,0.13},
{329,344,0.88,0.12},
{345,357,0.89,0.11},
{358,374,0.9,0.1},
{375,391,0.91,0.09},
{392,411,0.92,0.08},
{412,432,0.93,0.07},
{433,456,0.94,0.06},
{457,484,0.95,0.05},
{485,517,0.96,0.04},
{518,559,0.97,0.03},
{560,619,0.98,0.02},
{620,734,0.99,0.01},
{735,9999,1,0},
{-1,-1,-1,-1} };

// ---------------------------------------------------------------------- \\

// Ausgabetext generieren
bool CalcGenerateText ( const CALC_INPUT & inp, const CALC_RESULT & res, char * szBuffer, int iMaxBuffer )
{
  // berschrift und Hinweise
  LoadString ( globals().hInst, IDS_CALCTEXT + inp.iType , szBuffer, iMaxBuffer );

  // Eingegebene Daten
  if ( inp.iType == TYPE_DWZ )
    wsprintf ( szBuffer, "%sAlte Zahl:\t\t\t%i\r\nAlter:\t\t\t%i\r\nBisherige Berechnungen:\t%i\r\n\r\n", szBuffer, inp.iRo, inp.iAge, inp.iCount );
  if ( inp.iType == TYPE_ELO )
    wsprintf ( szBuffer, "%sAlte Zahl:\t\t\t%i\r\nBisherige gewertete Partien:\t%i\r\n\r\n", szBuffer, inp.iRo, inp.iCount );

  // Spiele
  for ( int i = 0; i < OPP_COUNT; ++i )
    if ( inp.iRatings [i] > 0 && inp.iRatings [i] < 4000 )
    {
      wsprintf ( szBuffer, "%sSpiel %i:\tgegen %i %s\r\n", szBuffer, i + 1, inp.iRatings [i], 
         inp.iScore [i] == 0 ? "verloren" : ( inp.iScore [i] == 1 ? "remis" : "gewonnen" ) ); 
    }

  // Berechnungsergebisse
  sprintf ( szBuffer, "%s\r\nGewertete Partien:\t%i\r\nPunkte:\t\t%g\r\nProzentual:\t%i%%\r\nGegnerschnitt:\t%i\r\n", 
            szBuffer, res.iN, res.dW, (int) (res.dW * 100 / res.iN), res.iRc );
  if ( res.dWe != CALC_NOTUSED && res.dWe != CALC_FALSE )
    sprintf ( szBuffer, "%sErwartete Punkte:\t%g\r\n", szBuffer, res.dWe );  
  if ( res.iRh != CALC_NOTUSED && res.iRh != CALC_FALSE )
    sprintf ( szBuffer, "%sLeistung:\t\t%i\r\n", szBuffer, res.iRh );  
  sprintf ( szBuffer, "%sNeue Wertung:\t%i\r\n", szBuffer, res.iRn );
  return true;  
}

// ---------------------------------------------------------------------- \\

int CalcRating ( const CALC_INPUT &input, CALC_RESULT * pr, int iReserved )
{
  // Berechnung von Ro, n, W, Rc
  pr->iRo = input.iRo;
  for ( int i = 0; i < OPP_COUNT; ++i )
  {
    if ( input.iRatings [i] > 0 && input.iRatings [i] < 4000 )
    {
      ++pr->iN;
      pr->dW += ( input.iScore [i] ) / 2.0;  
      pr->iRc += ( input.iRatings [i] );
    }
  }
  if ( pr->iN < 1 ) return 3;
  pr->iRc /= pr->iN;

  // Weitere Berechnungen je nach Typ
  switch ( input.iType )
  {
    // DWZ-Halbzahl
    case TYPE_FIRSTDWZ:
    {
      // Initialisierung
      pr->iRo = CALC_NOTUSED;
      pr->dWe = CALC_NOTUSED;
            
      // Rh = Rn berechnen
      pr->iRn = pr->iRh = CalcRh ( table, input, *pr );
      if ( pr->iRn < 800 ) pr->iRn = 700 + ( pr->iRh / 8 );
      return ERROR_SUCCESS;
    }
    
    // DWZ berechnen
    case TYPE_DWZ:
    { 
      // Fehler abfangen
      if ( input.iRo < 1 ) return 2;    
      if ( input.iAge < 1 ) return 4;
      if ( input.iCount < 1 ) return 5;

      // Gewinnerwartung We
      for ( int i = 0; i < OPP_COUNT; ++i )
        if ( input.iRatings [i] > 0 && input.iRatings [i] < 4000 )
          pr->dWe += CalcGetPD ( table, input.iRo, input.iRatings [i] );
                
      // Leistung berechnen
      pr->iRh = CalcRh ( table, input, *pr );
      
      // Entwicklungskeoffizienten berechnen
      double dfB = 1.0, dSBr = 0.0, dE;
      int iJ = 10;

      // J berechnen
      if ( input.iAge < 21 ) iJ = 5;
      if ( input.iAge > 25 ) iJ = 15;
      
      // Beschleunigungsfaktor
      if ( iJ == 5 && pr->dW >= pr->dWe )
        dfB = min ( 1.0, max ( 0.5, pr->iRo / 2000.0 ) );
 
      // Bremszuschlag
      if ( pr->iRo < 1300 && pr->dW <= pr->dWe )
        dSBr = pow ( CALC_E, (1300 - pr->iRo) / 150.0 ) - 1;
    
      // E und seine Grenzen 
      dE = max ( 5.0, pow ( pr->iRo / 1000.0, 4.0 ) + iJ * dfB + dSBr );
      if ( dSBr == 0 ) 
        dE = (int) min ( 150.0, min ( 5 * input.iCount, dE ) );  
      else
        dE = (int) min ( 30.0, dE );
      
      // Neue DWZ berechnen
      pr->iRn = (int) pr->iRo + 800.0 * ( pr->dW - pr->dWe ) / ( dE + pr->iN );      

      return ERROR_SUCCESS; 
    }
    
    // ELO Halbzahl berechnen
    case TYPE_FIRSTELO:
    {
      // Initialisierung
      int iD;
      pr->iRo = CALC_NOTUSED;
      pr->dWe = CALC_NOTUSED;
      pr->iRh = CALC_NOTUSED;          
                  
      // Wertungsdifferenz berechnen
      if ( pr->dW >= pr->iN / 2.0 )
        pr->iRn = pr->iRc + 25 * ( pr->dW - pr->iN / 2.0 );
      else
        pr->iRn = pr->iRc + CalcGetD ( table, pr->dW / pr->iN );     

      return ERROR_SUCCESS;
    }

    // ELO berechnen
    case TYPE_ELO:
    { 
      int iK, iRc2;
      
      // Fehler abfangen
      if ( input.iRo < 1 ) return 2;    
      if ( input.iCount < 1 ) return 5;

      // Rc2 wegen 350-Differenz berechnen
      iRc2 = 0;
      for ( int i = 0; i < OPP_COUNT; ++i )
        if ( input.iRatings [i] > 0 && input.iRatings [i] < 4000 )
        {
          if ( input.iRatings [i] < pr->iRo - 350 ) iRc2 += pr->iRo - 350;
          else if ( pr->iRo + 350 < input.iRatings [i] ) iRc2 += pr->iRo + 350;
          else iRc2 += input.iRatings [i];
        }
      iRc2 /= pr->iN;
      
      // Entwicklungskoeffizenten K berechnen
      if ( input.iCount <= 30 ) iK = 25;
      else if ( input.iRo >= 2400 || iReserved ) iK = 10;
      else iK = 15;

      // Gewinnerwartung We
      pr->dWe = pr->iN * CalcGetPD ( table, pr->iRo, iRc2 );
    
      // Leistung berechnen
      if ( pr->dW == 0.0 || pr->dW == pr->iN )
        pr->iRh = CALC_FALSE;
      else
        pr->iRh = pr->iRc + CalcGetD ( table, pr->dW / pr->iN );
            
      // Neue DWZ berechnen
      pr->iRn = (int) pr->iRo + iK * ( pr->dW - pr->dWe );      

      return ERROR_SUCCESS; 
    }
        
    default:
      return 1; 
  }
} 

// ---------------------------------------------------------------------- \\

int CalcRh ( CALC_ROW table [], const CALC_INPUT & input, const CALC_RESULT & result )
{      
  int iRp, iLast;
  double dWe;
  
  // Extremresultate
  if ( result.dW == 0.0 ) return result.iRc - 677;
  if ( result.dW == result.iN ) return result.iRc + 677;
  
  // Nherungswert
  iRp = result.iRc + CalcGetD ( table, result.dW / result.iN );
  
  // Iterationsverfahren: Verbesserte Leistung berechnen
  iLast = CALC_FALSE;
  do {
    iLast = iRp;
    dWe = 0.0;
    for ( int i = 0; i < OPP_COUNT; ++i )
      if ( input.iRatings [i] > 0 && input.iRatings [i] < 4000 )
        dWe += CalcGetPD ( table, iRp, input.iRatings [i] );
    iRp += CalcGetD ( table, ( result.dW - dWe ) / result.iN + 0.5 );
  } while  ( iLast != iRp );

  return iRp;
}     

// Gewinnwahrscheinlichkeit aus Tabelle lesen
double CalcGetPD ( CALC_ROW table [], int iRo, int iOpp )
{
  int iD = abs ( iRo - iOpp );
  for ( int i = 0; table [i].iMin >= 0; ++i )
  {
    if ( table [i].iMin <= iD && iD <= table [i].iMax )
      return (iRo > iOpp) ? table [i].dBigger : table [i].dSmaller;
  }
  return CALC_FALSE;
} 

// Wertungsdifferenz aus Tabelle lesen
int CalcGetD ( CALC_ROW table [], double dP )
{
  // Auf 2 Stellen runden
  dP = ((double)((int) (dP * 100.0))) / 100.0;
  
  for ( int i = 0; table [i].iMin >= 0; ++i )
  {
    // Sonderbehandlung fr besondere Werte
    if ( dP == 0.5 ) return 0;
    if ( dP == 0.0 ) return -677;
    if ( dP == 1.0 ) return 677;

    // Wenn Prozentualgewinn der Tabelle entspricht, Mittelwert zurckgeben
    if ( dP == table [i].dBigger )
      return ( table [i].iMin + table [i].iMax ) / 2;
    if ( dP == table [i].dSmaller )
      return - ( table [i].iMin + table [i].iMax ) / 2;
  }
  return CALC_FALSE;
} 

// ---------------------------------------------------------------------- \\

// Datei laden
int CalcLoad ( char * szFilename, CALC_INPUT * pcalc )
{
  ifstream in ( szFilename, ios::in | ios::binary );
  if ( in.rdstate () != ios::goodbit )
    return 6;
  in.read ( (char *) pcalc, sizeof ( CALC_INPUT ) );
  in.close ();
  
  // Welches Format?
  if ( pcalc->iSignatur == SIGNATUR )
    return ERROR_SUCCESS;
  else if ( pcalc->iSignatur == 1 )
  {
    // Altes Format
    CALC_INPUT_V1 import;

    ifstream in ( szFilename, ios::in | ios::binary );
    if ( in.rdstate () != ios::goodbit )
      return 6;
    in.read ( (char *) &import, sizeof ( CALC_INPUT_V1 ) );
    in.close ();

    pcalc->iSignatur = 1;
    pcalc->iType = TYPE_DWZ;
    pcalc->iRo = import.iRo;
    pcalc->iAge = import.iAge + 13;
    pcalc->iCount = 6;
    for ( int i = 0; i < 11; ++i )
    {
      pcalc->iScore [i] = import.dScore [i] * 2.0;
      pcalc->iRatings [i] = import.iRatings [i];
    }    

    return ERROR_SUCCESS;
  }
  else
    return 9;
}

// Datei speichern
int CalcSave ( char * szFilename, CALC_INPUT * pcalc )
{
  ofstream out ( szFilename, ios::out | ios::binary );
  if ( out.rdstate () != ios::goodbit )
    return 7;
  pcalc->iSignatur = SIGNATUR;
  out.write ( (char *) pcalc, sizeof ( CALC_INPUT ) );
  return ERROR_SUCCESS;
}

// ---------------------------------------------------------------------- \\
